import os
import sys
import argparse
import json
import importlib
import JinjaStaticRenderer as renderer
import Reader as Reader
import Core as Core
import Extensions as Extensions
import inspect

class Wrangler():

    defaults = {
        "generator_config": {
            "build_cache_file": "var/build.cache",
            "default_template": "base.j2",
            "templates_dir": "templates",
            "compiled_templates_file": "var/jinja",
            "compiled_templates_log": "var/jinja.log",
            "output_file_extension": "html",
            "output_dir": "www",
            "input_dir": "site",
            "data_format": "json",
            "ignore": [".", "_"],
            "site_vars":"site_vars",
            "verbose": "false",
            "force": "false",
            "nocache": "false",
            "item_class": "Page",
            "lib_path": "site/lib",
            "extensions": {
                "SiteMap": {
                    "options": {
                        "webroot":"/"
                    }
                },
                "CacheBuster": {
                    "options": {}
                },
                "FileInfo": {
                    "options": {
                        "directory": "www/assets",
                        "filetypes": ["css", "pdf"],
                        "webroot": "www"
                    }
                },
                "MooExtension": {
                    "options": {}
                }
            }
        },
        "site_vars": {
            "paths": {
                "css": "assets/css",
                "js": "assets/js",
                "assets": "assets",
                "images": "assets/images",
                "webroot": "www",
                "app": "app",
                "content": "content"
            }
        }
    }

    # Default config just loads the YAML config file
    config = {
        'config_path':'wrangler.json',
    }

    def __init__(self):
        return None

    def parse_args(self, args=None):
        d = '\033[32mProcess recursive directories of JSON files through Jinja2 templates. \033[0m'
        parser = argparse.ArgumentParser(description=d)
        parser.add_argument('input_dir',
                            help='\033[34mInput directory such as `site/content`\033[0m')
        parser.add_argument('output_dir',
                            help='\033[34mOutput directory such as `www`\033[0m')
        parser.add_argument("-t", "--templates_dir",
                            help='\033[34mTemplate directory such as `templates`\033[0m')
        parser.add_argument("-c", "--config",
                            help='\033[34mPath to `config.json`\033[0m')
        parser.add_argument("-o", "--output_file_extension",
                            help='\033[34mType of files to output such as `html`\033[0m')
        parser.add_argument("-d", "--data_format",
                            help='\033[34mType of files to match in input_dir, such as `json`\033[0m')
        parser.add_argument("-v", "--verbose", action="store_true",
                            help='\033[34mPrint all the plumbing\033[0m')
        parser.add_argument("-f", "--force", action="store_true",
                            help='\033[34mSmash out all the templates regardless of mtime\033[0m')
        parser.add_argument("-n", "--nocache", action="store_true",
                            help='\033[34mTurn off data persistence\033[0m')
        return parser.parse_args(args)


    def update_config(self, args):
        # If a config file is specified, load all config from there instead
        if (args.__dict__["config"] != None):
            self.config['config_path'] = args.__dict__["config"]


        self.config.update( self.defaults["generator_config"] )

        if os.path.exists( self.config['config_path'] ):
            userConfig = json.load( file(self.config['config_path']) )
            self.config.update( userConfig["generator_config"] )
        # else: 
            # userConfig = self.defaults

        # Apply settings from the config file (if present)
        # self.config.update( userConfig["generator_config"] )

        # Apply command line flags if set, ignore Nones.
        self.config.update((k, v) for k, v in args.__dict__.iteritems() if v is not None)
        
        # The site vars object is mapped to an item in the json object
        self.config["site_vars"] = userConfig[userConfig["generator_config"]["site_vars"]]



    def main(self, args=None):
        args = self.parse_args(args)

        self.update_config(args)

        if "lib_path" in self.config:
            self.load_classes(self.config["lib_path"])

        self._reporter = Core.Reporter(self.config)
        self._reader = Reader.Reader(self.config)
        self._writer = Core.Writer(self.config["output_dir"], self.config["output_file_extension"], self._reporter)
        self._renderer = renderer.JinjaStaticRenderer(self.config, self._reporter, self._writer)

        self._reporter.log("Digesting \"%s\" files from \"%s\"" % (self.config["data_format"], self.config["input_dir"]), "blue")
        self.graph = self._reader.fetch()


        # Set output paths before trying to render anything.
        # Two loops not so efficient... hmmm
        def set_output(node):
            if node.tag == 'file':
                cargo = node.get_cargo()
                if cargo:
                    cargo.set_output_path(self._writer.generate_output_path(cargo.relpath()))
            else:
                for child in node.children:
                    set_output(child)

        set_output(self.graph.tree())


        if "WranglerHooks" in sys.modules and hasattr(sys.modules["WranglerHooks"], "BeforeRender"):
            hook = sys.modules["WranglerHooks"].BeforeRender(self.config, self._renderer)
            hook.process(self.graph)

        self.process_extensions()
    

        rendered = 0

        for key, node in self.graph.all().items():
            if node.tag == 'file':
                # Leaky boat.
                cargo = node.get_cargo()
                if cargo:
                    cargo.set_parents(node.get_parents())
                    cargo.set_children(node.get_child_pages())
                    cargo.set_siblings(node.get_siblings())
                    cargo.set_parents_siblings(node.get_parents_siblings())
                    
                    _rel = cargo.get_related()

                    if _rel != None:
                        arr = self.graph.all()
                        related = []
                        for item in _rel:
                            _item = arr.get(item)
                            if _item:
                                related.append(_item)
                        
                        cargo.set_related_nodes(related)

                    if self._writer.save(self._renderer.render(cargo)):
                        rendered += 1

                    cargo.cleanup()
                    del cargo
    

        if "WranglerHooks" in sys.modules and hasattr(sys.modules["WranglerHooks"], "AfterRender"):
            hook = sys.modules["WranglerHooks"].AfterRender(self.config, self._renderer)
            hook.process(self.graph)
        
        if rendered > 0:
            self._reporter.log("Built %s of %s pages in \"%s\" directory " % (rendered, len(self.graph.all()) ,self.config["output_dir"]), "green")
            self._reporter.set_last_build_time()
        else:
             self._reporter.log("Hmm... nothing's changed in \"%s\" or \"%s\" since last time.\nIf you've removed pages, or included a template dynamically, try --force" % (self.config["input_dir"], self.config["templates_dir"]), "red")

    def load_classes(self, views): 
        classfiles = [os.path.join(dirpath, f)
            for dirpath, dirnames, files in os.walk(views)
            for f in files if f.endswith('.py')]

        path = list(sys.path)
        sys.path.insert(0, views)

        moduleNames = []


        for f in classfiles:
            moduleNames.append(os.path.basename(f.replace(".py", "")))

        if len(moduleNames) > 0:
            try:
                map(__import__, moduleNames);
            finally:
                # restore the syspath
                sys.path[:] = path 


    def process_extensions(self):
        """
        Load extensions from the core and import any from the site/lib directory too
        """
        extension_classes = inspect.getmembers(sys.modules["wrangler.Extensions"], inspect.isclass)

        if "lib_path" in self.config and "Extensions" in sys.modules:
            extension_classes = extension_classes + inspect.getmembers(sys.modules["Extensions"], inspect.isclass)

        if extension_classes:
            for extension_name, extension_config in self.config["extensions"].items():
                try:
                    ext = [(name, obj) for (name, obj) in extension_classes if name == extension_name]
                    
                    if ext[0]:
                        extension = ext[0][1](extension_config, self.graph)
                        self._reporter.verbose("Running extension %s > site.%s" % (ext[0][0], ext[0][0].lower()))
                        self.config["site_vars"][extension_name.lower()] = extension.run()
                except:
                    self._reporter.log("Couldn't load extension %s" % (extension_name), "red")


def start():
    sys.exit(Wrangler().main())

if __name__ == '__main__':
    start()



import pprint 
import time 
import os
import wrangler.Core as wrangler

class SiteMap(wrangler.Extension):
    def process_node(self, node):
        data = []
        for key in node.children:
            if key.tag == "dir":
                index = key.get_child("index")
                index_data = None
                if index:
                    cargo = index.get_cargo()
                    if cargo:   
                        index_data = cargo.get_properties()

                data.append({"data": index_data, "children": self.process_node(key)})
            else:
                if not key.is_index:
                    cargo = key.get_cargo()
                    if cargo:
                        data.append({"data": key.get_cargo().get_properties(), "children": {}})
        return data
        
    def run(self):
        self.webroot = self.config["options"]["webroot"]
        res = self.process_node(self.node_graph.tree())
        if "debug" in self.config["options"] and self.config["options"]["debug"] == True:
            p = pprint.PrettyPrinter(depth=12)
            p.pprint(res)
        return res


class CacheBuster(wrangler.Extension):
    def __init__(self, config={"options":{}}, node_graph=None):
        self.config = config
        
    def run(self):
        return int(time.time())


class FileInfo(wrangler.Extension):
    def __init__(self, config={"options":{"directory":"assets", "filetypes": ["css", "pdf"]}}, node_graph=None):
        self.config = config

    def run(self):
        assets = {}
        for root, dirs, files in os.walk(self.config["options"]["directory"]):
            files = [f for f in files if not f[0] == '.' and any(f.endswith(x) for x in self.config["options"]["filetypes"])]
            dirs[:] = [d for d in dirs if not d[0] == '.']
            for fn in files:
                path = os.path.join(root, fn)
                file_type = os.path.splitext(path)[1].replace(os.path.extsep, "")
                mtime = os.path.getmtime(path)
                size = os.stat(path).st_size # in bytes
                assets[path] = {
                    "path": path.replace(self.config["options"]["webroot"], ""),
                    "name": fn,
                    "size": size,
                    "type": file_type,
                    "mtime": mtime
                }
        return assets





import os
import subprocess
import shelve 
import sys 
import Core
import Parsers


class NodeGraph():
    def __init__(self):
        self._nodes = {}
        self._root = None

    def add(self, item):
        self._nodes[item.path] = item

    def root(self, item):
        self._root = item

    def tree(self):
        return self._root

    def all(self):
        return self._nodes


# Takes a directory and transforms the matching elements into Page instances.
class Reader():
    default_class = "Page"

    def __init__(self, config):
        self.input_dir = config['input_dir']
        self.ignore_files = config['ignore']
        self.data_format = config['data_format']
        self.nocache = config['nocache']
        self.config = config
        self.custom_default_module = self.check_custom_default()


    def check_custom_default(self):
        filename = "%s/%s.py" % (self.config["lib_path"], self.default_class)
        return os.path.exists(filename)


    def load_class(self, file_class):
        filename = "%s/%s.py" % (self.config["lib_path"], file_class)

        if not os.path.exists(filename):
            raise Exception("path %s doesn't exist in %s" % (file_class, self.config["lib_path"]))

        directory, module_name = os.path.split(filename)
        module_name = os.path.splitext(module_name)[0]

        path = list(sys.path)
        sys.path.insert(0, directory)

        try:
            module = __import__(module_name)
        finally:
            sys.path[:] = path # restore

        for name in dir(module):
            if module_name.split(".")[0] == name:
                return getattr(module, name)

        return Page
    

    def init_cache(self):
        if (self.config["nocache"]):
            try:
                os.remove("wrangler_cache.db")
            except OSError:
                pass

        return shelve.open("wrangler_cache") 
    
    def save_cache(self, shelf):
        try:
            shelf.sync()
        finally:
            shelf.close()


    def dir_as_tree(self, path, parent=None):
        """Recursive function that walks a directory and returns a tree
        of nested nodes.
        """

        basename = os.path.basename(path)
        node = Core.Node(basename, path, parent)
        self.graph.add(node)

        # Gather some more information on this path here
        # and write it to attributes
        if os.path.isdir(path):
            # Recurse
            node.tag = 'dir'
            for item in sorted(os.listdir(path)):
                item_path = os.path.join(path, item)

                # Ignore all the crufty things that get scraped, as well as hidden fies
                if not os.path.basename(item_path).startswith(".") and not ":" in item_path and not "%" in item_path and not "?" in item_path:
                    child_node = self.dir_as_tree(item_path, node)
                    node.add_child(child_node)
            return node
        else:
            node.tag = 'file'

            if basename == "index.%s" % (self.data_format):
                node.is_index = True
                
            return node


    def fetch(self):
        shelf = self.init_cache()
        parser = self.load_parser_by_format(self.data_format, self.input_dir, "")

        self.graph = NodeGraph()

        root_node = self.dir_as_tree(self.input_dir)        

        
        self.graph.root(root_node)
        
        for key, node in self.graph.all().items():
            if node.tag == "file":

                node.add_cargo(self.new_item(parser, shelf, node.path))
        
        self.save_cache(shelf)
        return self.graph


    def load_parser_by_format(self, data_format, input_dir, root):

        for parser in dir(Parsers):
            p =  getattr(Parsers, parser)
            if hasattr(p, "__bases__") and hasattr(p, "accepts"):
                if data_format == p.accepts:
                    return p(input_dir, root)

        raise Exception("No parser found for %s" % (data_format))


    def new_item(self, parser, shelf, filename):
        mtime = os.path.getmtime(filename)
        if (not shelf.has_key(filename)) or (shelf[filename].get_mtime() < mtime) or (self.nocache):
            
            # Check if custom class is set, otherwise make it a page... 
            page_data = parser.load(filename);

            if page_data:

                page_view = page_data["meta"]["view"]

                if (page_view != None and "lib_path" in self.config):
                    PageClass = self.load_class(page_view)
                elif self.custom_default_module:
                    PageClass = self.load_class(self.default_class)
                else:
                    PageClass = getattr(Core, self.default_class)

                new_page = PageClass(page_data, self.config)
                shelf[filename] = new_page
                print "\033[1;35mCaching \033[0m\033[2m%s\033[0m" % (filename)
                return new_page
        else:
            return shelf[filename]



import yaml
import json
import Core

class YamlParser(Core.Parser):
    accepts = "yaml"

    def interpret(self, file_contents):
        return yaml.load(file_contents)


class JsonParser(Core.Parser):
    accepts = "json"

    def interpret(self, file_contents):
        return json.load(file_contents)


import os
import markdown as md
import sys
import traceback
import glob
import Core
import utilities as util

from docutils.core import publish_parts
from jinja2 import Environment, FileSystemLoader, meta, BaseLoader, TemplateNotFound, ChoiceLoader, Template, Markup
from SilentUndefined import SilentUndefined
from MarkdownExtension import MarkdownExtension


def rst_filter(s):
    return Markup(publish_parts(source=s, writer_name='html')['body'])

def markdown_filter(value):
    """
    Jinja2 markdown filter with support for extensions.
    """
    output = value
    marked = md.Markdown(
        extensions=['extra','meta','sane_lists'],
        output_format='html5',
        safe_mode=False 
        )
    out = marked.convert(output)
    return out

def filter_template_dir(filename):
    if filename.startswith("."):
        return False;
    return True;


class JinjaStaticRenderer(Core.Renderer):

    def init(self):

        # Setup the env
        self.env = Environment(
            undefined=SilentUndefined,
            loader=FileSystemLoader(self.config['templates_dir']),
            extensions=[
                MarkdownExtension
            ],
            trim_blocks = True,
            lstrip_blocks = True
            )
        
        self.env.filters["markdown"] = markdown_filter
        self.env.filters['rst'] = rst_filter        
        self.template_trees = {}
        self.template_modified_times = {}
        
        var_path = os.path.dirname(self.config['compiled_templates_file'])

        self.reporter.verbose("Ensure directory exists: \033[34m%s\033[0m" % (var_path))
        util.ensure_dir(var_path)
    
        self.env.compile_templates(
            self.config['compiled_templates_file'],
            ignore_errors=False,
            filter_func=filter_template_dir 
            )

        self.reporter.verbose("Loading templates from: \033[34m%s\033[0m" % (self.config['templates_dir']))
        self.reporter.verbose("Compile templates to .zip: \033[32m%s\033[0m" % (self.config['compiled_templates_file']))


    def get_template_mtime(self, template):
        """
        Request the template's mtime. If we don't have it yet, cache it.
        """
        if not template in self.template_modified_times:
            self.template_modified_times[template] = os.path.getmtime("%s/%s" % (self.config['templates_dir'], template) )
        return self.template_modified_times[template]


    def get_referenced_templates(self, template):
        """
        Request the template's Abstract Syntax Tree so we can find other template references.
        Store the reference list in a dictionary
        """

        if not template in self.template_trees:
            try: 
                ast = self.env.parse(self.env.loader.get_source(self.env, template))
                self.template_trees[template] = list(meta.find_referenced_templates(ast))
            except:
                self.template_trees[template] = list()

            self.reporter.verbose("\033[34m%s\033[0m references templates \033[32m%s\033[0m" % (template, self.template_trees[template]))

        return self.template_trees[template]


    def render(self, item):
        template_name = item.get_template()
        template = self.load_template(template_name)

        if template and (self.should_render_item(item, template_name)):
            # self.reporter.verbose("\033[37m%s last built at: %s\033[0m" % (item.get_file_path(), item.get_modified_time()))
            return (template.render(data=item.get_content(), meta=item.get_metadata(), site=self.config["site_vars"]), item)

        return (False, item)


    def load_template(self, template_name):
        template_object = False
        try:
            template_object = self.env.get_template(template_name)
        except:
            print "\033[1;91mCouldn't parse `%s`. Check the template for errors \033[0;37m" % (template)
            
            if self.config["verbose"] == True:
                traceback.print_exc()
            print "\033[0m"
            return False
        finally:
            return template_object


    def should_render_item(self, item, template_name):
        force_render = 0
        last_build_time = self.reporter.get_last_build_time()
        output_path = item.get_output_path()

        if last_build_time == 0:
            return True

        config_modified_time = self.reporter.get_config_modified_time()
        
        if last_build_time == 0 or config_modified_time >= last_build_time:
            return True

        if self.config["force"] == True:
            return True

        if output_path and not os.path.exists(output_path):
            return True
        
        if item.get_modified_time() >= last_build_time:
            return True

        template_modified_time = 0

        if template_name:

            referenced_templates = self.get_referenced_templates(template_name)
            template_modified_time = self.get_template_mtime(template_name)
            
            if template_modified_time >= last_build_time:
                return True

            ref_template_mtime = 0

            for t in referenced_templates:
                time = self.get_template_mtime(t)
                if time >= last_build_time:
                    ref_template_mtime = time
                    return True

        return False
